library IEEE;

use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

entity dct is
  port (
    reset_n: in std_logic;
    clk: in std_logic;
    waitrequest : out std_logic;

    byteenable: in std_logic_vector(3 downto 0);
    begintransfer: in std_logic;
    chipselect: in std_logic;
    read: in std_logic;
    write: in std_logic;
    address: in std_logic_vector(18 downto 0);
    writedata: in std_logic_vector(31 downto 0);
    readdata : out std_logic_vector(31 downto 0)
  );
end dct;

architecture behav of dct is

type buf_type  is array (31 downto 0) of std_logic_vector (31 downto 0);
type buf_type0 is array (15 downto 0) of std_logic_vector (31 downto 0);   
type buf_type1 is array (7 downto 0) of std_logic_vector (31 downto 0);
type buf_type2 is array (3 downto 0) of std_logic_vector (31 downto 0);
type buf_type3 is array (1 downto 0) of std_logic_vector (31 downto 0);

signal pnt0  : buf_type0;
signal pnt1  : buf_type1;
signal pnt2  : buf_type2;
signal pnt3  : buf_type3;
signal pnt4  : std_logic_vector (31 downto 0);

signal buf1,buf2 : buf_type;
signal index, counter  : integer;
signal start,start_pre : std_logic;
signal end_flg     : std_logic;

signal add_a : buf_type;
signal add_b : buf_type;
signal add_result : buf_type;

signal sub_a : buf_type;
signal sub_b : buf_type;
signal sub_result : buf_type;

signal mult_a: std_logic_vector (31 downto 0);
signal mult_b: std_logic_vector (31 downto 0);
signal mult_result: std_logic_vector (31 downto 0);

signal i,j,m,n : integer;
signal operation_done : std_logic;

component add is
	port
	(
		dataa		: IN STD_LOGIC_VECTOR (31 DOWNTO 0);
		datab		: IN STD_LOGIC_VECTOR (31 DOWNTO 0);
		result		: OUT STD_LOGIC_VECTOR (31 DOWNTO 0)
	);
end component;

component sub IS
	PORT
	(
		dataa		: IN STD_LOGIC_VECTOR (31 DOWNTO 0);
		datab		: IN STD_LOGIC_VECTOR (31 DOWNTO 0);
		result		: OUT STD_LOGIC_VECTOR (31 DOWNTO 0)
	);
end component;

component mul_wrapper IS       
	PORT
	(
		dataa		: IN STD_LOGIC_VECTOR (31 DOWNTO 0);
		datab		: IN STD_LOGIC_VECTOR (31 DOWNTO 0);
		result64    : OUT STD_LOGIC_VECTOR (31 DOWNTO 0)    
        
     );
end component;

begin
waitrequest <= '0';

----------------------------------------------------
--decwin : buf is
--port map (
--  address => decwin_addr,
--  clock   => clk,
--  data    => decwin_wrdata,
--  wren    => decwin_wr,
--  q       => decwin_q
--);

--outbuf : buf is
--port map (
--  address => sndbuf_addr,
--  clock   => clk,
--  data    => sndbuf_wrdata,
--  wren    => sndbuf_wr,
--  q       => sndbuf_q
--);

-- decwin_addr <= address(9 downto 0);
-- decwin_wren <= decwin_wr or (write and address(18) and (not address(17)));
-- sndbuf_addr <= address(9 downto 0);
-- sndbuf_wren <= sndbuf_wr or (write and address(18) and address(17));

-----------------------------------------------------
--  Make 16 adders
-----------------------------------------------------
loop_add_inst:
for i in 0 to 16 generate 
	add_inst : add PORT MAP (
		dataa	 => add_a(i),
		datab	 => add_b(i),
		result	 => add_result(i)
	);
end generate;
-----------------------------------------------------
--  Make 16 subtractors
-----------------------------------------------------
loop_sub_inst:
for j in 0 to 16 generate
	sub_inst : sub PORT MAP (
		dataa	 => sub_a(j),
		datab	 => sub_b(j),
		result	 => sub_result(j)
	);
end generate;

-----------------------------------------------------
--  Make only 1 multiplier
--  64bits*64bits=128bits, only the lower 64bits of mult_result are used
-----------------------------------------------------
mult_inst : mul_wrapper PORT MAP (
		dataa	 => mult_a,
		datab	 => mult_b,
		result64	 => mult_result   
       
	);

-----------------------------------------------------
--  save previous value of "start" signal
--  used to start counter
-----------------------------------------------------
process (clk,reset_n)
begin
  if reset_n='0' then
	start_pre <= '0';
  elsif (clk'event and clk='1') then
	start_pre <= start;
  end if;
end process;


-----------------------------------------------------
--  counter begin from "start" signal
--  end @ 89
--  generate "end-flg" signal 
--  end-flg <= '1'
-----------------------------------------------------
operation_done <= '1' when counter = 89 else '0';

process (clk,reset_n)
begin
  if reset_n='0' then
    start <= '0';
  elsif (clk'event and clk='1') then
    if(chipselect='1' and write='1' and index=64) then
        start<=writedata(0);
    elsif ( operation_done = '1' ) then
        start <= '0'; 
    end if;
  end if;
end process;

process (clk,reset_n)
begin
  if reset_n='0' then
    counter <= 0;
    end_flg <= '0';
  elsif (clk'event and clk='1') then
    if (start_pre='0' and start='1') then
      counter <= 1;
      end_flg <= '0';
    elsif (operation_done = '1' ) then
      counter <= 89;
      end_flg <= '1';
    elsif ( start='1') then
      counter <= counter +1;
    end if;
  end if;
end process;


-----------------------------------------------------
--  transfer address from avalon to integer
-----------------------------------------------------
index <= conv_integer(address(7 downto 0));

-----------------------------------------------------
--  output loop
--  output "end_flg" to avalon bus @ index 64
-----------------------------------------------------

process(chipselect, read, index, buf1, buf2)
begin
  if (chipselect='1' and read='1') then
      for i in 0 to 31 loop
        if(index = i)   then readdata <= buf1(i)(31 downto 0); end if;
      end loop;
      for i in 0 to 31 loop
        if(index = i +32) then readdata <= buf2(i)(31 downto 0); end if;
      end loop;
      if(index = 64) then
        readdata <= "0000000000000000000000000000000" & end_flg;
      end if;
  else
    readdata <= (others => '0');
  end if;
end process;
-----------------------------------------------------
--  reading inputs and state machine
-----------------------------------------------------
process (clk,reset_n)
begin
  if reset_n='0' then
     for i in 0 to 31 loop
       buf1(i) <= (others=>'0');
       buf2(i) <= (others=>'0');
     end loop;
  elsif (clk'event and clk='1') then
    if ( chipselect='1' and write='1' ) then
        for i in 0 to 31 loop
          if (index = i ) then
            buf1(i)(31 downto 0) <= writedata;
            buf2(i)(31 downto 0) <= writedata;
          end if;
        end loop;
    elsif (start_pre='1' and start='1') then
----------------------------------------------------------------------------
-- pnt0
----------------------------------------------------------------------------
    if (counter = 1) then
      for i in 0 to 15 loop
        buf1(i)    <= add_result(i);
        buf1(i+16) <= sub_result(i);    
      end loop;
    end if;
	for i in 0 to 15 loop
	  if (counter = i+2) then buf1(i+16) <= mult_result; end if;
	end loop;
-----------------------------------------------------------------------------
-- pnt1
-----------------------------------------------------------------------------
    if(counter = 18) then
      for i in 0 to 7 loop
        buf2(i     )<= add_result(i);
        buf2(i+8   )<= sub_result(i);
        buf2(i  +16)<= add_result(i+8);
        buf2(i+8+16)<= sub_result(i+8);
      end loop;
    end if;
    for i in 0 to 7 loop
       if (counter = i+19) then buf2(i+8   ) <= mult_result; end if;
    end loop;
    for i in 0 to 7 loop
       if (counter = i+27) then buf2(i+8+16) <= mult_result; end if;
    end loop;
-----------------------------------------------------------------------------
-- pnt2
-----------------------------------------------------------------------------
    if(counter = 35) then
      for i in 0 to 3 loop
        buf1(i       )<= add_result(i);
        buf1(i+4     )<= sub_result(i);
        buf1(i  +8   )<= add_result(i+4);
        buf1(i+4+8   )<= sub_result(i+4);
        buf1(i    +16)<= add_result(i+8);
        buf1(i+4  +16)<= sub_result(i+8);
        buf1(i  +8+16)<= add_result(i+12);
        buf1(i+4+8+16)<= sub_result(i+12);
      end loop;
    end if;
    for i in 0 to 3 loop
       if (counter = i+36) then buf1(i+4     ) <= mult_result; end if;
    end loop;
    for i in 0 to 3 loop
       if (counter = i+40) then buf1(i+4+8   ) <= mult_result; end if;
    end loop;
    for i in 0 to 3 loop
       if (counter = i+44) then buf1(i+4  +16) <= mult_result; end if;
    end loop;
    for i in 0 to 3 loop
       if (counter = i+48) then buf1(i+4+8+16) <= mult_result; end if;
    end loop;
-----------------------------------------------------------------------------
-- pnt3
-----------------------------------------------------------------------------
    if(counter = 52) then
      for i in 0 to 1 loop
        buf2(i         )<= add_result(i);
        buf2(i+2       )<= sub_result(i);
        buf2(i  +4     )<= add_result(i+2);
        buf2(i+2+4     )<= sub_result(i+2);
        buf2(i    +8   )<= add_result(i+4);
        buf2(i+2  +8   )<= sub_result(i+4);
        buf2(i  +4+8   )<= add_result(i+6);
        buf2(i+2+4+8   )<= sub_result(i+6);
        buf2(i      +16)<= add_result(i+8);
        buf2(i+2    +16)<= sub_result(i+8);
        buf2(i  +4  +16)<= add_result(i+10);
        buf2(i+2+4  +16)<= sub_result(i+10);
        buf2(i    +8+16)<= add_result(i+12);
        buf2(i+2  +8+16)<= sub_result(i+12);
        buf2(i  +4+8+16)<= add_result(i+14);
        buf2(i+2+4+8+16)<= sub_result(i+14);
      end loop;
    end if;
    for i in 0 to 1 loop
       if (counter = i+53) then buf2(i+2   ) <= mult_result; end if;
    end loop;
    for i in 0 to 1 loop
       if (counter = i+55) then buf2(i+2+4 ) <= mult_result; end if;
    end loop;
    for i in 0 to 1 loop
       if (counter = i+57) then buf2(i+2+8 ) <= mult_result; end if;
    end loop;
    for i in 0 to 1 loop
       if (counter = i+59) then buf2(i+2+12) <= mult_result; end if;
    end loop;
    for i in 0 to 1 loop
       if (counter = i+61) then buf2(i+2+16) <= mult_result; end if;
    end loop;
    for i in 0 to 1 loop
       if (counter = i+63) then buf2(i+2+20) <= mult_result; end if;
    end loop;
    for i in 0 to 1 loop
       if (counter = i+65) then buf2(i+2+24) <= mult_result; end if;
    end loop;
    for i in 0 to 1 loop
       if (counter = i+67) then buf2(i+2+28) <= mult_result; end if;
    end loop;
-----------------------------------------------------------------------------
-- pnt4
-----------------------------------------------------------------------------
    if(counter = 69) then
      for i in 0 to 15 loop
        buf1(i*2  )<= add_result(i);
        buf1(i*2+1)<= sub_result(i);
      end loop;
    end if;
    for i in 0 to 15 loop
      if(counter = i+70) then buf1(i*2+1) <= mult_result; end if;
    end loop;
-----------------------------------------------------------------------------
    if(counter = 86) then
      for i in 0 to 7 loop
        buf1(2+i*4) <= add_result(i);
      end loop;
    end if;
    if(counter = 87) then
      for i in 0 to 3 loop
        buf1(4+i*8) <= add_result(i*3);
        buf1(6+i*8) <= add_result(i*3+1);
        buf1(5+i*8) <= add_result(i*3+2);
      end loop;
    end if;
    if(counter = 88) then
      for i in 0 to 1 loop
        buf1(8+i*16)  <= add_result(i*7);
        buf1(12+i*16) <= add_result(i*7+1);
        buf1(10+i*16) <= add_result(i*7+2);
        buf1(14+i*16) <= add_result(i*7+3);
        buf1(9+i*16)  <= add_result(i*7+4);
        buf1(13+i*16) <= add_result(i*7+5);
        buf1(11+i*16) <= add_result(i*7+6);
      end loop;
    end if;
    
------------------------------------------------------------------------------
    end if; --start_pre start
  end if;   --clock'event
end process;

-----------------------------------------------------
--  assign input value to adder/sub/multiplier
-----------------------------------------------------
process(counter,add_a,add_b,sub_a,sub_b,mult_a,mult_b)
begin
loop_add_assign: for m in 0 to 31 loop
    add_a(m) <= (others=>'0');
    add_b(m) <= (others=>'0');
end loop;    
loop_sub_assign: for n in 0 to 31 loop
    sub_a(n) <= (others=>'0');
    sub_b(n) <= (others=>'0');
end loop;
mult_a <= (others=>'0');
mult_b <= (others=>'0');

----------------------------------------------------------------------------
-- pnt0
----------------------------------------------------------------------------
  if(counter = 1) then
    for i in 0 to 15 loop
      add_a(i) <= buf1(i);
      add_b(i) <= buf1(31-i);
      sub_b(i) <= buf2(16+i);
      sub_a(i) <= buf2(15-i);
    end loop;
  end if;
  for i in 0 to 15 loop
    if(counter = i+2) then
      mult_a <= buf1(i+16);
      mult_b <= pnt0(15-i);
    end if;
  end loop;
-----------------------------------------------------------------------------
-- pnt1
-----------------------------------------------------------------------------
  if (counter = 18) then
    for i in 0 to 7 loop
      add_a(i) <= buf1(i);
      add_b(i) <= buf1(15-i);
      sub_b(i) <= buf1(8+i);
      sub_a(i) <= buf1(7-i);

      add_a(i+8) <= buf1(0+i +16);
      add_b(i+8) <= buf1(15-i+16);
      sub_b(i+8) <= buf1(7-i +16);
      sub_a(i+8) <= buf1(8+i +16);
    end loop;
  end if;
  for i in 0 to 7 loop
    if(counter = i+19) then
      mult_a <= buf2(i+8);
      mult_b <= pnt1(7-i);
    end if;
  end loop;
  for i in 0 to 7 loop
    if(counter = i+27) then
      mult_a <= buf2(i+8+16);
      mult_b <= pnt1(7-i);
    end if;
  end loop;
-----------------------------------------------------------------------------
-- pnt2
-----------------------------------------------------------------------------
  if (counter = 35) then
    for i in 0 to 3 loop
      add_a(i)    <= buf2(i);
      add_b(i)    <= buf2(7-i);
      sub_b(i)    <= buf2(4+i);
      sub_a(i)    <= buf2(3-i);

      add_a(i+4)  <= buf2(0+i +8);
      add_b(i+4)  <= buf2(7-i +8);
      sub_b(i+4)  <= buf2(3-i +8);
      sub_a(i+4)  <= buf2(4+i +8);

      add_a(i+8)  <= buf2(0+i +16);
      add_b(i+8)  <= buf2(7-i +16);
      sub_b(i+8)  <= buf2(4+i +16);
      sub_a(i+8)  <= buf2(3-i +16);

      add_a(i+12) <= buf2(i   +24);
      add_b(i+12) <= buf2(7-i +24);
      sub_b(i+12) <= buf2(3-i +24);
      sub_a(i+12) <= buf2(4+i +24);
    end loop;
  end if;
  for i in 0 to 3 loop
    if(counter = i+36) then
      mult_a <= buf1(i+4);
      mult_b <= pnt2(3-i);
    end if;
  end loop;
  for i in 0 to 3 loop
    if(counter = i+40) then
      mult_a <= buf1(i+4+8);
      mult_b <= pnt2(3-i);
    end if;
  end loop;
  for i in 0 to 3 loop
    if(counter = i+44) then
      mult_a <= buf1(i+4+16);
      mult_b <= pnt2(3-i);
    end if;
  end loop;
  for i in 0 to 3 loop
    if(counter = i+48) then
      mult_a <= buf1(i+4+8+16);
      mult_b <= pnt2(3-i);
    end if;
  end loop;
-----------------------------------------------------------------------------
-- pnt3
-----------------------------------------------------------------------------
  if (counter = 52) then
    for i in 0 to 1 loop
      add_a(i)    <= buf1(i);
      add_b(i)    <= buf1(3-i);
      sub_b(i)    <= buf1(2+i);
      sub_a(i)    <= buf1(1-i);

      add_a(i+2)  <= buf1(i   +4);
      add_b(i+2)  <= buf1(3-i +4);
      sub_b(i+2)  <= buf1(1-i +4);
      sub_a(i+2)  <= buf1(2+i +4);

      add_a(i+4)  <= buf1(0+i +8);
      add_b(i+4)  <= buf1(3-i +8);
      sub_b(i+4)  <= buf1(2+i +8);
      sub_a(i+4)  <= buf1(1-i +8);

      add_a(i+6)  <= buf1(0+i +12);
      add_b(i+6)  <= buf1(3-i +12);
      sub_b(i+6)  <= buf1(1-i +12);
      sub_a(i+6)  <= buf1(2+i +12);

      add_a(i+8)  <= buf1(0+i +16);
      add_b(i+8)  <= buf1(3-i +16);
      sub_b(i+8)  <= buf1(2+i +16);
      sub_a(i+8)  <= buf1(1-i +16);

      add_a(i+10) <= buf1(0+i +20);
      add_b(i+10) <= buf1(3-i +20);
      sub_b(i+10) <= buf1(1-i +20);
      sub_a(i+10) <= buf1(2+i +20);

      add_a(i+12) <= buf1(i   +24);
      add_b(i+12) <= buf1(3-i +24);
      sub_b(i+12) <= buf1(2+i +24);
      sub_a(i+12) <= buf1(1-i +24);

      add_a(i+14) <= buf1(0+i +28);
      add_b(i+14) <= buf1(3-i +28);
      sub_b(i+14) <= buf1(1-i +28);
      sub_a(i+14) <= buf1(2+i +28);
    end loop;
  end if;
  for i in 0 to 1 loop
    if(counter = i+53) then
      mult_a <= buf2(i+2);
      mult_b <= pnt3(1-i);
    end if;
  end loop;
  for i in 0 to 1 loop
    if(counter = i+55) then
      mult_a <= buf2(i+2+4);
      mult_b <= pnt3(1-i);
    end if;
  end loop;
  for i in 0 to 1 loop
    if(counter = i+57) then
      mult_a <= buf2(i+2+8);
      mult_b <= pnt3(1-i);
    end if;
  end loop;
  for i in 0 to 1 loop
    if(counter = i+59) then
      mult_a <= buf2(i+2+12);
      mult_b <= pnt3(1-i);
    end if;
  end loop;
  for i in 0 to 1 loop
    if(counter = i+61) then
      mult_a <= buf2(i+2+16);
      mult_b <= pnt3(1-i);
    end if;
  end loop;
  for i in 0 to 1 loop
    if(counter = i+63) then
      mult_a <= buf2(i+2+20);
      mult_b <= pnt3(1-i);
    end if;
  end loop;
  for i in 0 to 1 loop
    if(counter = i+65) then
      mult_a <= buf2(i+2+24);
      mult_b <= pnt3(1-i);
    end if;
  end loop;
  for i in 0 to 1 loop
    if(counter = i+67) then
      mult_a <= buf2(i+2+28);
      mult_b <= pnt3(1-i);
    end if;
  end loop;
-----------------------------------------------------------------------------
-- pnt4
-----------------------------------------------------------------------------
  if(counter = 69) then
    for i in 0 to 7 loop
      add_a(i*2)   <= buf2(i*4);
      add_b(i*2)   <= buf2(i*4+1);
      sub_b(i*2)   <= buf2(i*4+1);
      sub_a(i*2)   <= buf2(i*4);

      add_a(i*2+1)  <= buf2(i*4+2);
      add_b(i*2+1)  <= buf2(i*4+3);
      sub_b(i*2+1)  <= buf2(i*4+2);
      sub_a(i*2+1)  <= buf2(i*4+3);
    end loop;
  end if;
  for i in 0 to 15 loop
    if(counter = i+70) then
      mult_a <= buf1(i*2+1);
      mult_b <= pnt4;
    end if;
  end loop;
-----------------------------------------------------------------------------
  if(counter = 86) then
    for i in 0 to 7 loop
      add_a(i) <= buf1(2+i*4);
      add_b(i) <= buf1(3+i*4);
    end loop;
  end if;
  if(counter = 87) then
    for i in 0 to 3 loop
      add_a(i*3) <= buf1(4+i*8);
      add_b(i*3) <= buf1(6+i*8);
      add_a(i*3+1) <= buf1(6+i*8);
      add_b(i*3+1) <= buf1(5+i*8);
      add_a(i*3+2) <= buf1(5+i*8);
      add_b(i*3+2) <= buf1(7+i*8);
    end loop;
  end if;
  if(counter = 88) then
    for i in 0 to 1 loop
      add_a(i*7) <= buf1(8+i*16);
      add_b(i*7) <= buf1(12+i*16);
      add_a(i*7+1) <= buf1(12+i*16);
      add_b(i*7+1) <= buf1(10+i*16);
      add_a(i*7+2) <= buf1(10+i*16);
      add_b(i*7+2) <= buf1(14+i*16);
      add_a(i*7+3) <= buf1(14+i*16);
      add_b(i*7+3) <= buf1(9+i*16);
      add_a(i*7+4) <= buf1(9+i*16);
      add_b(i*7+4) <= buf1(13+i*16);
      add_a(i*7+5) <= buf1(13+i*16);
      add_b(i*7+5) <= buf1(11+i*16);
      add_a(i*7+6) <= buf1(11+i*16);
      add_b(i*7+6) <= buf1(15+i*16);
    end loop;
  end if;
  if(counter = 89) then
    
  end if;
end process;

-- cosine table
pnt0(0)(31 downto 0)<=conv_std_logic_vector(16403,32);
pnt0(1)(31 downto 0)<=conv_std_logic_vector(16563,32);
pnt0(2)(31 downto 0)<=conv_std_logic_vector(16890,32);
pnt0(3)(31 downto 0)<=conv_std_logic_vector(17401,32);
pnt0(4)(31 downto 0)<=conv_std_logic_vector(18124,32);
pnt0(5)(31 downto 0)<=conv_std_logic_vector(19101,32);
pnt0(6)(31 downto 0)<=conv_std_logic_vector(20398,32);
pnt0(7)(31 downto 0)<=conv_std_logic_vector(22112,32);
pnt0(8)(31 downto 0)<=conv_std_logic_vector(24396,32);
pnt0(9)(31 downto 0)<=conv_std_logic_vector(27503,32);
pnt0(10)(31 downto 0)<=conv_std_logic_vector(31869,32);
pnt0(11)(31 downto 0)<=conv_std_logic_vector(38320,32);
pnt0(12)(31 downto 0)<=conv_std_logic_vector(48633,32);
pnt0(13)(31 downto 0)<=conv_std_logic_vector(67429,32);
pnt0(14)(31 downto 0)<=conv_std_logic_vector(111660,32);  
pnt0(15)(31 downto 0)<=conv_std_logic_vector(333906,32);  

pnt1(0)(31 downto 0)<=conv_std_logic_vector(16463,32);
pnt1(1)(31 downto 0)<=conv_std_logic_vector(17121,32);
pnt1(2)(31 downto 0)<=conv_std_logic_vector(18577,32);
pnt1(3)(31 downto 0)<=conv_std_logic_vector(21195,32);
pnt1(4)(31 downto 0)<=conv_std_logic_vector(25826,32);
pnt1(5)(31 downto 0)<=conv_std_logic_vector(34756,32);
pnt1(6)(31 downto 0)<=conv_std_logic_vector(56441,32);
pnt1(7)(31 downto 0)<=conv_std_logic_vector(167154,32);

pnt2(0)(31 downto 0)<=conv_std_logic_vector(16704,32);
pnt2(1)(31 downto 0)<=conv_std_logic_vector(19704,32);
pnt2(2)(31 downto 0)<=conv_std_logic_vector(29490,32);
pnt2(3)(31 downto 0)<=conv_std_logic_vector(83981,32);

pnt3(0)(31 downto 0)<=conv_std_logic_vector(17733,32);
pnt3(1)(31 downto 0)<=conv_std_logic_vector(42813,32);

pnt4   (31 downto 0)<=conv_std_logic_vector(23170,32);

end behav;




